unit AVHintEdit;

{ Heavily based on TWAIHintListBox freeware component, from the article
  'Dropping Hints: Part II', found in Delphi Informant Magazine,
  http://www.delphizine.com/features/2000/01/di200001ma_f/di200001ma_d.asp,
  by Motty Adler.  Adapted by Andre Viol

  If you modify something for better, please, let me know - andre@viol.net

  TAVHintEdit is just like a TEdit but with an extra property: HintText.
  if HintText is set to True, then, when there is a text that is wider than the
  Edit width, a hint will popup when the mouse is over the Edit. That's all! }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, StdCtrls;

type

  TPaintLineHintWindow = class(THintWindow)
  private
    //procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
      { You can uncomment this procedure and its related code to override the
        default border of the hint.  If so, comment out the public procedures
        Paint and CreateParams, and their related code, since the feature these
        procedures provide are not necessary anymore. }
  public
    procedure Paint; override;
    procedure CreateParams(var Params: TCreateParams); override;
  end;

  TAVHintEdit = class(TEdit)
  private
    { Private declarations }
    FShowing: Boolean;
      {True if the hint window is visible and activated}
    FSelectAll: Boolean;
      {True if SelectAll must be called when TEdit is to be focused}
    FWider: Boolean;
      {True if TEdit text is wider then TEdit width}
    FWindow: TForm;
      {Instance of window on which to operate}
      {Since the TEdit component has no Canvas, we use the Form canvas to which
       the TEdit belongs.}
    FHintText: Boolean;
      {True if the extra hint text must be shown if FWider is True}
      { To avoid having a permanent hint over some dialog, if you call a modal
        dialog from the same form where there is an AVHintEdit component,
        set HintText property temporarily to False; otherwise, if the modal
        dialog takes a certain time to open and, if, in the meantime, the user
        moves the mouse over the AVHintEdit, it is possible to have a 'frozen'
        hint over the modal dialog, when it pops up.}
    FOnMouseExit: TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    procedure ReleaseIt;
    procedure SetHintText(const Value: Boolean);
    procedure SetOnMouseEnter(const Value: TNotifyEvent);
    procedure SetOnMouseExit(const Value: TNotifyEvent);
  protected
    { Protected declarations }
    procedure CMMouseEnter(var message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var message: TMessage); message CM_MOUSELEAVE;
  public
    { Public declarations }
    HintWin: TPaintLineHintWindow;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DefaultHandler(var message); override;
  published
    { Published declarations }
    property OnMouseEnter:TNotifyEvent read FOnMouseEnter write SetOnMouseEnter;
    property OnMouseExit:TNotifyEvent read FOnMouseExit write SetOnMouseExit;
    property HintText:Boolean read FHintText write SetHintText;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('AViol', [TAVHintEdit]);
end;

{ TPaintLineHintWindow }
{-----------------------------------------------------------------------------}
{procedure TPaintLineHintWindow.WMNCPaint(var Message: TMessage);
// this code was kindly provided by Peter Below (TeamB), on Borland Newsgroup
begin
  message.result :=
    DefWindowProc(handle, message.msg, message.wparam, message.lparam);
end;}

{-----------------------------------------------------------------------------}
procedure TPaintLineHintWindow.CreateParams(var Params: TCreateParams);
{ We need to remove the border created on the Windows API-level }
{ when the window is created. }
begin
  inherited CreateParams(Params);
  // remove the original border because it is NOT all in black, it uses
  // clBtnFace and this makes the appearance of the border ugly, when over
  // a form, IMHO, since most forms are clBtnFaced already.
  Params.Style := Params.Style and not ws_Border;
end;

procedure TPaintLineHintWindow.Paint;
begin
  inherited; // draw the original hint
  with Canvas do begin  // then, draw a new border
    Pen.Width := 1;
    Pen.Color := clBtnShadow; // clBlack is the windows default;
    Polyline([Point(0,0),Point(ClipRect.Right-1,0),
              Point(ClipRect.Right-1,ClipRect.Bottom-1),
              Point(0,ClipRect.Bottom-1),Point(0,0)]);
  end;
end;

{ TAVHintEdit }
{-----------------------------------------------------------------------------}
procedure TAVHintEdit.CMMouseEnter(var message: TMessage);
begin
  inherited;
  if Assigned(FOnMouseEnter) then fonmouseenter(Self);
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.CMMouseLeave(var message: TMessage);
begin
  inherited;
  ReleaseIt;
  If Assigned(FOnMouseExit) then fonmouseexit(Self);
end;

{-----------------------------------------------------------------------------}
constructor TAVHintEdit.Create(AOwner: TComponent);
var
  TheForm: TComponent;       {The form that ultimately owns this component}
begin
  inherited Create(AOwner);
  HintWin := TPaintLineHintWindow.Create(Self);
  HintWin.Color := clInfoBk;
  HintWin.Canvas.Font.Color := clInfoText;

  { The idea/code below was borrowed from TPJWdwState freeware component, by
   peter.johnson@openlink.org.  See http://www.pjsoft.contactbox.co.uk/ }

  {Check if owner is a form, if not, move up tree of owners until we either find
  a form or a nil reference}
  TheForm := AOwner;
  while (TheForm <> nil) and not (TheForm is TForm) do
    TheForm := TheForm.Owner;
  {Record form window reference appropriately}
  if TheForm <> nil then FWindow := TheForm as TForm
                    else FWindow := nil;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.DefaultHandler(var message);
var
  MsgT: TMsg;
begin
  // convert message to TMsg
  MsgT.message := TMessage(message).Msg;
  inherited;
  if HintWin = nil then Exit;

  //if it's WM_Mousemove then we don't want to hide the hint window
  if (HintWin.IsHintMsg(MsgT) and (MsgT.message <> WM_MouseMove)) then
    ReleaseIt;
end;

{-----------------------------------------------------------------------------}
destructor TAVHintEdit.Destroy;
begin
  //HintWin.Free;
  inherited;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if FSelectAll and FWider and Focused then begin
    SelectAll;
    FSelectAll := False;
  end;
  inherited;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
{ Since this was originally coded for a TListBox, where it makes sense being
  this way, and here it was adapted for a TEdit component, I think this could
  be better off inside the OnEnter event, but, since I'm naturally born lazy,
  I made only the minor modifications needed.  Someone can surely do a better
  job... }
var
  Rct: TRect;
begin
  if FWider and not Focused then begin
    FSelectAll := True;
    FWider := False;
  end;

  if not(HintText) then begin  //Item hinting is off
    ReleaseIt;
    inherited;
    Exit;
  end;

  // Check if there is an owning window - get out if not
  // If HintEdit is focused, do not show the tooltip to let user edit it
  // Also, if Application has no focus, why should we care about showing a tip?
  if Focused or (FWindow = nil) or (Trim(Text) = '') or
    not Application.Active then Exit;

  // the line that follows is important, otherwise the next line will calculate
  // the wrong TextWidth via FWindows' Canvas - Thanks Peter Below (TeamB),
  // for this handy tip that I found in Borland Newsgroup
  FWindow.Canvas.Font := Font;
  // Check if the Text is wider then the Width
  if FWindow.Canvas.TextWidth(Text) > Width - 6 then begin
    FWider := True;
    if FShowing then Exit;
    // Get the default item coordinates
    Rct := ClientRect;
    // Stretch it to fit the whole edit text
    Rct.Right := Rct.Left + FWindow.Canvas.TextWidth(Text) + 5;
    // Fine tune theses values for appearance
    Rct.Top := Rct.Top - 1;//2;
    Rct.Bottom := Rct.Bottom - 4;//2;
    Rct.Left := Rct.Left - 1;//2;
    // now convert to screen coordinates so that THintWindow can use them
    Rct.TopLeft := ClientToScreen(Rct.TopLeft);
    Rct.BottomRight := ClientToScreen(Rct.BottomRight);
    // And show it!
    HintWin.ActivateHint(Rct, Text);
    FShowing := True;
  end
  else
    ReleaseIt;

  inherited;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.SetHintText(const Value: Boolean);
begin
  FHintText := Value;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.SetOnMouseEnter(const Value: TNotifyEvent);
begin
  FOnMouseEnter := Value;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.SetOnMouseExit(const Value: TNotifyEvent);
begin
  FOnMouseExit := Value;
end;

{-----------------------------------------------------------------------------}
procedure TAVHintEdit.ReleaseIt;
begin
  HintWin.ReleaseHandle;
  FShowing := False;
end;

end.